@@ -21,10 +21,13 @@
  * Boston, MA 02111-1307, USA.
 */
 
+#include <config.h>
+
 #include <config.h>
 #include <math.h>
 #include <string.h>
 #include <stdio.h>
+#include <glib/gi18n-lib.h>
 #include "tasklist.h"
 #include "window.h"
 #include "class-group.h"
@@ -127,6 +130,9 @@ struct _WnckTask
   gdouble glow_start_time;
   
   guint button_glow;
+  
+  guint row;
+  guint col;
 };
 
 struct _WnckTaskClass
@@ -217,6 +223,8 @@ static gboolean wnck_task_get_needs_attention (WnckTask *task);
 
 static char      *wnck_task_get_text (WnckTask *task, gboolean icon_text);
 static GdkPixbuf *wnck_task_get_icon (WnckTask *task);
+static gint       wnck_task_compare_alphabetically (gconstpointer  a,
+                                                    gconstpointer  b);
 static gint       wnck_task_compare  (gconstpointer  a,
 				      gconstpointer  b);
 static void       wnck_task_update_visible_state (WnckTask *task);
@@ -294,52 +302,6 @@ static gpointer tasklist_parent_class;
  */
 static GSList *tasklist_instances;
 
-/**
- * make_gtk_label_bold.
- *
- * Switches the font of label to a bold equivalent.
- * @label: The label.
- **/
-static void
-make_gtk_label_bold (GtkLabel *label)
-{
-  PangoFontDescription *font_desc;
-
-  font_desc = pango_font_description_new ();
-
-  pango_font_description_set_weight (font_desc,
-                                     PANGO_WEIGHT_BOLD);
-
-  /* This will only affect the weight of the font, the rest is
-   * from the current state of the widget, which comes from the
-   * theme or user prefs, since the font desc only has the
-   * weight flag turned on.
-   */
-  gtk_widget_modify_font (GTK_WIDGET (label), font_desc);
-
-  pango_font_description_free (font_desc);
-}
-
-static void
-make_gtk_label_normal (GtkLabel *label)
-{
-  PangoFontDescription *font_desc;
-
-  font_desc = pango_font_description_new ();
-
-  pango_font_description_set_weight (font_desc,
-                                     PANGO_WEIGHT_NORMAL);
-
-  /* This will only affect the weight of the font, the rest is
-   * from the current state of the widget, which comes from the
-   * theme or user prefs, since the font desc only has the
-   * weight flag turned on.
-   */
-  gtk_widget_modify_font (GTK_WIDGET (label), font_desc);
-
-  pango_font_description_free (font_desc);
-}
-
 static GType
 wnck_task_get_type (void) G_GNUC_CONST;
 
@@ -704,11 +666,6 @@ wnck_tasklist_init (WnckTasklist *tasklist)
   atk_obj = gtk_widget_get_accessible (widget);
   atk_object_set_name (atk_obj, _("Window List"));
   atk_object_set_description (atk_obj, _("Tool to switch between visible windows"));
-
-  tasklist_instances = g_slist_append (tasklist_instances, tasklist);
-  g_slist_foreach (tasklist_instances,
-		   (GFunc) wnck_tasklist_update_lists,
-		   NULL);
 }
 
 static void
@@ -756,11 +713,6 @@ wnck_tasklist_finalize (GObject *object)
   WnckTasklist *tasklist;
 
   tasklist = WNCK_TASKLIST (object);
-  
-  tasklist_instances = g_slist_remove (tasklist_instances, tasklist);
-  g_slist_foreach (tasklist_instances,
-		   (GFunc) wnck_tasklist_update_lists,
-		   NULL);
 
   if (tasklist->priv->free_icon_loader_data != NULL)
     (* tasklist->priv->free_icon_loader_data) (tasklist->priv->icon_loader_data);
@@ -1271,6 +1223,8 @@ const int *
 wnck_tasklist_get_size_hint_list (WnckTasklist  *tasklist,
 				  int           *n_elements)
 {
+  g_return_val_if_fail (n_elements != NULL, NULL);
+
   *n_elements = tasklist->priv->size_hints_len;
   return tasklist->priv->size_hints;
 }
@@ -1295,6 +1249,7 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   GList *ungrouped_class_groups;
   WnckTask *win_task;
   GList *visible_tasks = NULL;
+  GList *windows_sorted = NULL;
   int grouping_limit;
   
   tasklist = WNCK_TASKLIST (widget);
@@ -1332,6 +1287,10 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
       if (g_list_length (class_group_task->windows) > 1)
 	{
 	  visible_tasks = g_list_prepend (visible_tasks, class_group_task);
+
+          /* Sort */
+          class_group_task->windows = g_list_sort (class_group_task->windows,
+                                                   wnck_task_compare_alphabetically);
 	  
 	  /* Hide all this group's windows */
 	  l = class_group_task->windows;
@@ -1413,6 +1372,24 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
       gtk_widget_size_allocate (task->button, &child_allocation);
       gtk_widget_set_child_visible (GTK_WIDGET (task->button), TRUE);
 
+      if (task->type != WNCK_TASK_STARTUP_SEQUENCE)
+        {
+          GList *ll;
+          
+          /* Build sorted windows list */
+          if (g_list_length (task->windows) > 1)
+            windows_sorted = g_list_concat (windows_sorted,
+                                           g_list_copy (task->windows));
+          else
+            windows_sorted = g_list_append (windows_sorted, task);
+          task->row = row;
+          task->col = col;
+          for (ll = task->windows; ll; ll = ll->next)
+            {
+              WNCK_TASK (ll->data)->row = row;
+              WNCK_TASK (ll->data)->col = col;
+            }
+        }
       i++;
       l = l->next;
     }
@@ -1421,7 +1398,9 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   wnck_tasklist_update_icon_geometries (tasklist, visible_tasks);
   
   g_list_free (visible_tasks);
-  
+  g_list_free (tasklist->priv->windows);
+  tasklist->priv->windows = windows_sorted;
+ 
   GTK_WIDGET_CLASS (tasklist_parent_class)->size_allocate (widget, allocation);
 }
 
@@ -1443,7 +1422,10 @@ wnck_tasklist_realize (GtkWidget *widget)
   
   (* GTK_WIDGET_CLASS (tasklist_parent_class)->realize) (widget);
 
-  wnck_tasklist_update_lists (tasklist);
+  tasklist_instances = g_slist_append (tasklist_instances, tasklist);
+  g_slist_foreach (tasklist_instances,
+		   (GFunc) wnck_tasklist_update_lists,
+		   NULL);
 }
 
 static void
@@ -1459,6 +1441,11 @@ wnck_tasklist_unrealize (GtkWidget *widget)
 #endif
   
   (* GTK_WIDGET_CLASS (tasklist_parent_class)->unrealize) (widget);
+  
+  tasklist_instances = g_slist_remove (tasklist_instances, tasklist);
+  g_slist_foreach (tasklist_instances,
+		   (GFunc) wnck_tasklist_update_lists,
+		   NULL);
 }
 
 static void
@@ -1654,6 +1641,86 @@ wnck_tasklist_set_screen (WnckTasklist *tasklist,
   wnck_tasklist_connect_screen (tasklist, screen);
 }
 
+static gboolean
+wnck_tasklist_scroll_cb (WnckTasklist *tasklist,
+                         GdkEventScroll *event,
+                         gpointer user_data)
+{
+  /* use the fact that tasklist->priv->windows is sorted is sorted
+   * see wnck_tasklist_size_allocate() */
+  GList *window;
+  gint row = 0;
+  gint col = 0;
+
+  window = g_list_find (tasklist->priv->windows,
+                        tasklist->priv->active_task);
+  if (window)
+    {
+      row = WNCK_TASK (window->data)->row;
+      col = WNCK_TASK (window->data)->col;
+    }
+  else
+    if (tasklist->priv->activate_timeout_id)
+      /* There is no active_task yet, but there will be one after the timeout.
+       * It occurs if we change the active task too fast. */
+      return TRUE;
+
+  switch (event->direction)
+    {
+      case GDK_SCROLL_UP:
+        if (!window)
+          window = g_list_last (tasklist->priv->windows);
+        else
+          window = window->prev;
+      break;
+
+      case GDK_SCROLL_DOWN:
+        if (!window)
+          window = tasklist->priv->windows;
+        else
+          window = window->next;
+      break;
+
+      case GDK_SCROLL_LEFT:
+        if (!window)
+          window = g_list_last (tasklist->priv->windows);
+        else
+          {
+            /* Search the first window on the previous colomn at same row */
+            while (window && (WNCK_TASK(window->data)->row != row ||
+                              WNCK_TASK(window->data)->col != col-1))
+              window = window->prev;
+            /* If no window found, select the first one */
+            if (!window)
+              window = tasklist->priv->windows;
+          }
+      break;
+
+      case GDK_SCROLL_RIGHT:
+        if (!window)
+          window = tasklist->priv->windows;
+        else
+          {
+            /* Search the first window on the next colomn at same row */
+            while (window && (WNCK_TASK(window->data)->row != row ||
+                              WNCK_TASK(window->data)->col != col+1))
+              window = window->next;
+            /* If no window found, select the last one */
+            if (!window)
+              window = g_list_last (tasklist->priv->windows);
+          }
+      break;
+
+      default:
+        g_assert_not_reached ();
+    }
+
+  if (window)
+    wnck_tasklist_activate_task_window (window->data, event->time);
+
+  return TRUE;
+}
+
 GtkWidget*
 wnck_tasklist_new (WnckScreen *screen)
 {
@@ -1667,6 +1734,13 @@ wnck_tasklist_new (WnckScreen *screen)
 
   wnck_tasklist_set_screen (tasklist, screen);
 
+  /* callback when there is a scroll-event for switching to the next window  */
+  g_signal_connect_object (G_OBJECT (tasklist),
+                           "scroll-event",
+                           G_CALLBACK (wnck_tasklist_scroll_cb),
+                           G_OBJECT (tasklist),
+                           0);
+  
   return GTK_WIDGET (tasklist);
 }
 
@@ -1748,7 +1822,8 @@ tasklist_include_window_ignoring_skip_taskbar (WnckTasklist *tasklist,
   if (active_workspace == NULL)
     return TRUE;
 
-  if (active_workspace != wnck_window_get_workspace (win))
+  if (active_workspace != wnck_window_get_workspace (win) &&
+      !wnck_window_or_transient_needs_attention (win))
     return FALSE;
 
   if (!wnck_workspace_is_virtual (active_workspace))
@@ -1978,7 +2053,7 @@ wnck_tasklist_active_window_changed (WnckScreen   *screen,
   /* FIXME: check for group modal window */
   for (active_window = wnck_screen_get_active_window (screen);
        active_window && !active_task;
-       active_window = wnck_window_get_transient(active_window))
+       active_window = wnck_window_get_transient (active_window))
     active_task = g_hash_table_lookup (tasklist->priv->win_hash,
                                        active_window);
 
@@ -2348,7 +2423,7 @@ wnck_task_popup_menu (WnckTask *task,
       text = wnck_task_get_text (win_task, FALSE);
       menu_item = gtk_image_menu_item_new_with_label (text);
       if (wnck_task_get_needs_attention (win_task)) 
-        make_gtk_label_bold (GTK_LABEL (GTK_BIN (menu_item)->child));
+        _make_gtk_label_bold (GTK_LABEL (GTK_BIN (menu_item)->child));
       g_free (text);
       
       pixbuf = wnck_task_get_icon (win_task);
@@ -2711,12 +2786,12 @@ wnck_task_update_visible_state (WnckTask *task)
       gtk_label_set_text (GTK_LABEL (task->label), text);
       if (wnck_task_get_needs_attention (task))
         {
-          make_gtk_label_bold ((GTK_LABEL (task->label)));
+          _make_gtk_label_bold ((GTK_LABEL (task->label)));
           wnck_task_queue_glow (task);
         }
       else
         {
-          make_gtk_label_normal ((GTK_LABEL (task->label)));
+          _make_gtk_label_normal ((GTK_LABEL (task->label)));
           wnck_task_stop_glow (task);
         }
       g_free (text);
@@ -2750,10 +2825,11 @@ wnck_task_state_changed (WnckWindow     *window,
       (changed_mask & WNCK_WINDOW_STATE_DEMANDS_ATTENTION) ||
       (changed_mask & WNCK_WINDOW_STATE_URGENT))
     {
-      WnckTask *win_task;
+      WnckTask *win_task = NULL;
 
-      win_task = g_hash_table_lookup (tasklist->priv->win_hash,
-						window);
+      /* FIXME: Handle group modal dialogs */
+      for (; window && !win_task; window = wnck_window_get_transient (window))
+        win_task = g_hash_table_lookup (tasklist->priv->win_hash, window);
 
       if (win_task)
 	{
@@ -2761,7 +2837,9 @@ wnck_task_state_changed (WnckWindow     *window,
 
 	  wnck_task_update_visible_state (win_task);
 	  
-	  class_group_task = g_hash_table_lookup (tasklist->priv->class_group_hash, win_task->class_group);
+	  class_group_task =
+            g_hash_table_lookup (tasklist->priv->class_group_hash,
+                                 win_task->class_group);
 
 	  if (class_group_task)
 	    wnck_task_update_visible_state (class_group_task);
@@ -2976,7 +3054,7 @@ wnck_task_create_widgets (WnckTask *task, GtkReliefStyle relief)
 
   if (wnck_task_get_needs_attention (task))
     {
-      make_gtk_label_bold ((GTK_LABEL (task->label)));
+      _make_gtk_label_bold ((GTK_LABEL (task->label)));
       wnck_task_queue_glow (task);
     }
 
@@ -3125,6 +3203,25 @@ wnck_task_expose (GtkWidget        *widget,
   return FALSE;
 }
 
+static gint
+wnck_task_compare_alphabetically (gconstpointer a,
+                                  gconstpointer b)
+{
+  char *text1;
+  char *text2;
+  gint  result;
+
+  text1 = wnck_task_get_text (WNCK_TASK (a), FALSE);
+  text2 = wnck_task_get_text (WNCK_TASK (b), FALSE);
+
+  result= g_utf8_collate (text1, text2);
+
+  g_free (text1);
+  g_free (text2);
+
+  return result;
+}
+
 static gint
 compare_class_group_tasks (WnckTask *task1, WnckTask *task2)
 {
@@ -3158,8 +3255,8 @@ wnck_task_compare (gconstpointer  a,
       pos1 = wnck_window_get_sort_order (task1->window);
       break;
     case WNCK_TASK_STARTUP_SEQUENCE:
-      pos1 = G_MAXINT;
-      break;
+      pos1 = G_MAXINT; /* startup sequences are sorted at the end. */
+      break;           /* Changing this will break scrolling.      */
     }
 
   switch (task2->type)
